/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 David N. Welton.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        David N. Welton (http://www.dedasys.com/davidw/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "sourcecache" and "David N. Welton" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    davidw@apache.org.
 *
 * 5. Products derived from this software may not be called "sourcecache",
 *    nor may "sourcecache" appear in their name, without prior written
 *    permission of the author(s).
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR(S) OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ==================================================================== */

/* This idea is inspired by Paolo Brutti's work on mod_dtcl. */

#include <tcl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

struct cacheinfo {
    int cachesize;
    int cachefree;
    Tcl_HashTable *cache;
    char **cachelist;
};

static int GetFile(Tcl_Interp *interp, char *filename, off_t filesize, Tcl_Obj *script)
{
    int result = 0;
    /* Taken, in part, from tclIOUtil.c out of the Tcl distribution,
       and modified */

    Tcl_Channel chan = Tcl_OpenFileChannel(interp, filename, "r", 0644);
    if (chan == (Tcl_Channel) NULL)
    {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "couldn't read file \"", filename,
			 "\": ", Tcl_PosixError(interp), (char *) NULL);
	return TCL_ERROR;
    }

    result = Tcl_ReadChars(chan, script, filesize, 1);
    if (result < 0)
    {
	Tcl_Close(interp, chan);
	Tcl_AppendResult(interp, "couldn't read file \"", filename,
			 "\": ", Tcl_PosixError(interp), (char *) NULL);
	return TCL_ERROR;
    }

    if (Tcl_Close(interp, chan) != TCL_OK)
	return TCL_ERROR;
    
    return TCL_OK;
}

static int SourceCacheFile(Tcl_Interp *interp, struct cacheinfo *cacheinfo, char *filename)
{
    char *hashkey = NULL;
    int isnew = 0;
    struct stat statPtr;
    Tcl_HashEntry *entry;
    Tcl_Obj *script;

    hashkey = Tcl_Alloc(strlen(filename) + 18); /* 18 should be enough
                                                   space to hold a 64
                                                   bit time_t printed
                                                   as hex */
    if (Tcl_Stat(filename, &statPtr) != 0)
	return TCL_ERROR;
    sprintf(hashkey, "%s%lx", filename, statPtr.st_mtime);
    entry = Tcl_CreateHashEntry(cacheinfo->cache, hashkey, &isnew);
    if (isnew)
    {
	script = Tcl_NewObj();
	Tcl_IncrRefCount(script);
	if (GetFile(interp, filename, statPtr.st_size, script) == TCL_ERROR)
	{
	    Tcl_Free(hashkey);
	    Tcl_DecrRefCount(script);
	    return TCL_ERROR;
	}
	
	if (cacheinfo->cachesize > 0)
	{
	    Tcl_SetHashValue(entry, (ClientData)script);
	    	    
	    if (cacheinfo->cachefree) {
		cacheinfo->cachelist[-- cacheinfo->cachefree ] = strdup(hashkey);
	    } else {
		int last = cacheinfo->cachesize - 1;
		Tcl_HashEntry *delEntry;
		delEntry = Tcl_FindHashEntry(cacheinfo->cache, cacheinfo->cachelist[last]);

		Tcl_DecrRefCount((Tcl_Obj *)Tcl_GetHashValue(delEntry));
		Tcl_DeleteHashEntry(delEntry);

		Tcl_Free(cacheinfo->cachelist[last]);
		memmove((cacheinfo->cachelist) + 1, cacheinfo->cachelist, sizeof(char *) * (cacheinfo->cachesize - 1));
		cacheinfo->cachelist[0] = strdup(hashkey);
	    }
	}
    } else {
	script = Tcl_GetHashValue(entry);	
    }
    Tcl_Free(hashkey);
    return Tcl_EvalObjEx(interp, script, 0);
}

static int SourceCacheInfo(Tcl_Interp *interp, struct cacheinfo *cacheinfo)
{
    Tcl_Obj *retlist;
    
    retlist = Tcl_NewObj();
    Tcl_IncrRefCount(retlist);
    Tcl_ListObjAppendElement(interp, retlist, Tcl_NewStringObj("cachesize", -1));
    Tcl_ListObjAppendElement(interp, retlist, Tcl_NewIntObj(cacheinfo->cachesize));

    Tcl_ListObjAppendElement(interp, retlist, Tcl_NewStringObj("cachefree", -1));
    Tcl_ListObjAppendElement(interp, retlist, Tcl_NewIntObj(cacheinfo->cachefree));

    Tcl_ListObjAppendElement(interp, retlist, Tcl_NewStringObj("hashstats", -1));
    Tcl_ListObjAppendElement(interp, retlist, 
			     Tcl_NewStringObj(Tcl_HashStats(cacheinfo->cache), -1));

    Tcl_SetObjResult(interp, retlist);
    return TCL_OK;
}

static int SourceCacheGetCache(Tcl_Interp *interp, struct cacheinfo *cacheinfo)
{
    char *hashkey;
    Tcl_Obj *data;
    Tcl_HashEntry *entry;
    Tcl_HashSearch search;
    Tcl_Obj *retlist;

    retlist = Tcl_NewObj();
    Tcl_IncrRefCount(retlist);

    entry = Tcl_FirstHashEntry(cacheinfo->cache, &search);
    while (entry != NULL)
    {
	data = Tcl_GetHashValue(entry);
	hashkey = Tcl_GetHashKey(cacheinfo->cache, entry);
	Tcl_ListObjAppendElement(interp, retlist, Tcl_NewStringObj(hashkey, -1));
	Tcl_ListObjAppendElement(interp, retlist, data);
	entry = Tcl_NextHashEntry(&search);
    }

    Tcl_SetObjResult(interp, retlist);
    return TCL_OK;
}


static int SourceCache(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    struct cacheinfo *cacheinfo = clientData;    
    char *command = NULL;
 
    if (objc < 2)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "file filename | info | getcache");
	return TCL_ERROR;
    }

    command = Tcl_GetString(objv[1]);
    if (!strcmp(command, "file"))
    {
	if (objc != 3) 
	    Tcl_WrongNumArgs(interp, 2, objv, "filename");
	else 
	    return SourceCacheFile(interp, cacheinfo, Tcl_GetString(objv[2]));	
    } else if (!strcmp(command, "info")) {
	if (objc != 2) 
	    Tcl_WrongNumArgs(interp, 2, objv, "info");
	else 
	    return SourceCacheInfo(interp, cacheinfo);	
    } else if (!strcmp(command, "getcache")) {
	if (objc != 2) 
	    Tcl_WrongNumArgs(interp, 2, objv, "info");
	else 
	    return SourceCacheGetCache(interp, cacheinfo);	
    } else {
	Tcl_WrongNumArgs(interp, 1, objv, "file filename | info | getcache");
    }
    return TCL_ERROR;
}

static void DeleteSourceCache(ClientData clientData)
{
    struct cacheinfo *cacheinfo;
    Tcl_Obj *data;
    Tcl_HashEntry *entry;
    Tcl_HashSearch search;

    cacheinfo = clientData;
    Tcl_Free((char *)(cacheinfo->cachelist));
    entry = Tcl_FirstHashEntry(cacheinfo->cache, &search);
    while (entry != NULL)
    {
	data = Tcl_GetHashValue(entry);
	Tcl_DecrRefCount(data);
	Tcl_DeleteHashEntry(entry);
	entry = Tcl_NextHashEntry(&search);
    }
    Tcl_DeleteHashTable(cacheinfo->cache);
}

static int MakeSourceCache(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    static int cachenum = 0;
    int cachesize = 0;
    struct cacheinfo *cacheinfo;
    char *cmdname;
    if (objc != 2)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "cache size");
	return TCL_ERROR;
    }
    cachenum ++;
    if (Tcl_GetIntFromObj(interp, objv[1], &cachesize) != TCL_OK)
	return TCL_ERROR;
    cacheinfo = (struct cacheinfo *)Tcl_Alloc(sizeof(struct cacheinfo));
    cacheinfo->cachesize = cachesize;
    cacheinfo->cachefree = cachesize;
    cacheinfo->cachelist = (char **)Tcl_Alloc(sizeof (char*) * cachesize);
    memset(cacheinfo->cachelist, 0, cachesize * sizeof(char *));
    cacheinfo->cache = (Tcl_HashTable *)Tcl_Alloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(cacheinfo->cache, TCL_STRING_KEYS);

    cmdname = Tcl_Alloc(20);
    sprintf(cmdname, "sourcecache%d", cachenum);
    Tcl_CreateObjCommand(interp, cmdname, SourceCache,
			 cacheinfo, DeleteSourceCache);
    Tcl_AppendResult(interp, cmdname, NULL);

    return TCL_OK;
}


int Sourcecache_Init(Tcl_Interp *interp)
{
    Tcl_CreateObjCommand(interp, "makesourcecache", MakeSourceCache,
			 NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_PkgProvide(interp, "sourcecache", "1.0");
    
    return TCL_OK;
}
